"""`metrics` module contains useful functions for image quality assessment."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/91_metrics.ipynb.

# %% auto 0
__all__ = ['epsilon', 'mse', 'rmse', 'nmse', 'nrmse', 'psnr', 'ssim', 'print_metrics']

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 4
import numpy as np
from prettytable import PrettyTable

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 5
from skimage.metrics import peak_signal_noise_ratio, structural_similarity
from typing import Optional

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 6
epsilon = 1e-10

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 8
def mse(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted) 
) -> np.ndarray:        # MSE (mean squared error)
    "Mean Squared Error (MSE). Broadcasting rules will be applied if different dimensions."
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if gt.ndim != pred.ndim: print("Warning: gt's and pred's dims are mismatched!")
    if gt.ndim <2: print("Warning: gt's dims < 2")
    if pred.ndim <2: print("Warning: pred's dims < 2")
    return np.array(np.mean((gt - pred) ** 2))

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 12
def rmse(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted) 
) -> np.ndarray:        # RMSE (root-mean-square error)
    "Root-Mean-Square Error (RMSE). Broadcasting rules will be applied if different dimensions."
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if gt.ndim != pred.ndim: print("Warning: gt's and pred's dims are mismatched!")
    if gt.ndim <2: print("Warning: gt's dims < 2")
    if pred.ndim <2: print("Warning: pred's dims < 2")
    return np.array(np.sqrt(np.mean((gt - pred) ** 2)))

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 16
def nmse(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted) 
) -> np.ndarray:        # NMSE (normalized mean-square error)
    "Normalized Mean-Square Error (NMSE). Broadcasting rules will be applied if different dimensions."
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if gt.ndim != pred.ndim: print("Warning: gt's and pred's dims are mismatched!")
    if gt.ndim <2: print("Warning: gt's dims < 2")
    if pred.ndim <2: print("Warning: pred's dims < 2")
    return np.array(np.linalg.norm(gt - pred) ** 2 / np.linalg.norm(gt) ** 2)


# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 20
def nrmse(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted) 
) -> np.ndarray:        # NRMSE
    "Normalized Root-Mean-Square Error (NRMSE). Broadcasting rules will be applied if different dimensions."
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if gt.ndim != pred.ndim: print("Warning: gt's and pred's dims are mismatched!")
    if gt.ndim <2: print("Warning: gt's dims < 2")
    if pred.ndim <2: print("Warning: pred's dims < 2")
    return np.array(np.sqrt(np.linalg.norm(gt - pred) ** 2 / np.linalg.norm(gt) ** 2))


# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 25
def psnr(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted)
    drange:float=None   # distance between min & max possible values
) -> np.ndarray:        # PSNR (peak signal-to-noise ratio)
    "Peak Signal-to-Noise Ratio (PSNR)"
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if gt.ndim != pred.ndim: print("Warning: gt's and pred's dims are mismatched!")
    if gt.ndim <2: print("Warning: gt's dims < 2")
    if pred.ndim <2: print("Warning: pred's dims < 2")
    if drange is None:
        drange = gt.max() - gt.min()
    return np.array(peak_signal_noise_ratio(gt, pred, data_range=drange))

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 28
def ssim(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted)
    drange: float=None  # distance between min & max values
) -> np.ndarray:        # SSIM
    "Structural Similarity Index Metric (SSIM)"
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if not gt.ndim == pred.ndim:
        raise ValueError("Ground truth dimensions does not match pred.")
    if not gt.ndim == 2:
        raise ValueError("Ground truth dimensions has to be 2.")
    drange = (gt.max()-gt.min()) if drange is None else drange
    return np.array(structural_similarity(gt, pred, data_range=drange))

# %% ../nbs/P4_HPD_cg/91_metrics.ipynb 31
def print_metrics(
    gt: np.ndarray,     # first image (ground truth)
    pred: np.ndarray,   # second image (predicted)
):
    "Print MSE, NMSE, RMSE, NRMSE, PSNR, and SSIM"
    gt, pred = np.squeeze(gt), np.squeeze(pred)
    if not gt.ndim == pred.ndim:
        raise ValueError("Ground truth dimensions does not match pred.")
    if not gt.ndim == 2:
        raise ValueError("Ground truth dimensions has to be 2.")
    
    # set the table up
    ptab = PrettyTable(['Metrics', 'Values'])
    # add rows
    ptab.add_row(['MSE', f"{mse(gt, pred):1.3e}"])
    ptab.add_row(['NMSE', f"{nmse(gt, pred):1.3e}"])
    ptab.add_row(['RMSE', f"{rmse(gt, pred):1.3e}"])
    ptab.add_row(['NRMSE', f"{nrmse(gt, pred):1.3e}"])
    ptab.add_row(['PSNR', f"{psnr(gt, pred):.5}"])
    ptab.add_row(['SSIM', f"{ssim(gt, pred):.5}"])
    # display
    print(ptab)



"""`optimizer` module contains the steepest descent and conjugate gradient methods."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/95_optim.ipynb.

# %% auto 0
__all__ = ['frobenius_norm', 'naive_steepest_descent', 'steepest_descent', 'conjugate_gradient']

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 3
import numpy as np
import h5py
import matplotlib.pyplot as plt

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 4
from .plot import *
from .metrics import *
from .math import *
from .io import *
import cgsense2023 as cgs2003
from .mri import *

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 6
def frobenius_norm(
    A:np.ndarray,   # input matrix A 
)->float:           # Frobenius norm of matrix A
    "Frobenius norm of a matrix A defined as the square root of the sum of the absolute squares of its elements"
    return np.sqrt(np.sum(np.square(np.abs(A))))

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 16
def naive_steepest_descent(
    mriop,                  # MRI operator NuFFT, etc.
    guess:np.ndarray,       # Initial Guess of Zeros
    data:np.ndarray,        # Dens_compensated kspace
    iters:int=10,           # Number of Iterations
    lr:float=0.4,           # Learning Rate
    tol:float=0.0,          # Tolerance to terminate
    lambd:float=0.0,        # Tikhonov Regularization Param
    kfilter:bool=True,      # flag to turn kspace filter True/False,
    dstore: bool=False,     # flag to store recon at every iter
    ref:np.ndarray=None,    # Ground truth reference image
)-> tuple:                  # return (final_im, res_norms, ref_norms)
    "naive gradient descent method"
    
    # number of readouts point & image dimension
    nreads, image_dim = data.shape[-1], guess.shape[-1]
    # We start with an initial guess of zeros..
    xs = []
    x = guess
    # calculate b (lhs)
    b = mriop.operator_rhs(data)
    # residual == gradient w/o regularization
    res = mriop.operator_lhs(x) - b
    # residual norms
    res_norms = []
    # reference norms
    if ref is not None: ref_norms = []
    for i in range(iters):
        # regularization
        gamma = res + lambd * x
        # update
        x = x - lr * gamma
        xs.append(x)
        # residual
        res = mriop.operator_lhs(x) - b
        delta = frobenius_norm(res)
        res_norms.append(delta)
        # reference norms
        if ref is not None: ref_norms.append(frobenius_norm(x-ref))
        if delta < tol:
            print("Converged after %i iterations to %1.6e" % (i + 1, delta))
            break
        if not np.mod(i, 1):
            print("Residuum at iter %i : %1.6e" % (i + 1, delta), end='\r')
    
    xs = np.array(xs)      
    if kfilter: xs = kspace_filter(xs, nreads, image_dim)
    
    if ref is not None:
        return xs if dstore else xs[-1], res_norms, ref_norms
    else:
        return xs if dstore else xs[-1], res_norms

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 19
def steepest_descent(
    mriop,                  # MRI operator NuFFT, etc.
    guess:np.ndarray,       # Initial Guess of Zeros
    data:np.ndarray,        # Dens_compensated kspace
    iters:int=10,           # Number of Iterations
    tol:float=0.0,          # Tolerance to terminate
    lambd:float=0.0,        # Tikhonov Regularization Param
    kfilter:bool=True,      # flag to turn kspace filter True/False
    dstore: bool=False,     # flag to store recon at every iter
    ref:np.ndarray=None,    # Ground truth reference image
)-> tuple:                  # return (final_im, res_norms, ref_norms)
    "naive gradient descent method"
    # number of readouts point & image dimension
    nreads, image_dim = data.shape[-1], guess.shape[-1]
    # We start with an initial guess of zeros..
    xs = []
    x = guess
    # calculate b (lhs)
    b = mriop.operator_rhs(data)
    # residual norms
    res_norms = []
    # res_norms.append(frobenius_norm(mriop.operator_lhs(x)-b))
    # reference norms
    if ref is not None: ref_norms = []
    for i in range(iters):
        # gamma - gradient
        gamma = mriop.operator_lhs(x) - b
        # H-gamma
        Hgamma = mriop.operator_lhs(gamma)
        # alpha - optimal learning rate
        alpha = np.vdot(gamma, gamma) / (np.vdot(gamma, Hgamma))
        # update
        x = x - alpha * (gamma + lambd*x)
        xs.append(x)
        # residual
        delta = frobenius_norm(mriop.operator_lhs(x) - b)
        res_norms.append(delta)
        # reference norms
        if ref is not None: ref_norms.append(frobenius_norm(x-ref))
        if delta < tol:
            print("Converged after %i iterations to %1.6e" % (i + 1, delta))
            break
        if not np.mod(i, 1):
            print("Residuum at iter %i : %1.6e" % (i + 1, delta), end='\r')
    xs = np.array(xs)
    if kfilter: xs = kspace_filter(xs, nreads, image_dim)
    if ref is not None:
        return xs if dstore else xs[-1], res_norms, ref_norms
    else:
        return xs if dstore else xs[-1], res_norms

# %% ../nbs/P4_HPD_cg/95_optim.ipynb 22
def conjugate_gradient(
    mriop,                  # MRI operator NuFFT, etc.
    guess:np.ndarray,       # Initial Guess of Zeros
    data:np.ndarray,        # Dens_compensated kspace
    iters:int=10,           # Number of Iterations
    tol:float=0.0,          # Tolerance to terminate
    lambd:float=0.0,        # Tikhonov Regularization Param
    kfilter:bool=True,      # flag to turn kspace filter True/False
    dstore: bool=False,     # flag to store recon at every iter
    ref:np.ndarray=None,    # Ground truth reference image
)-> tuple:                  # return (final_im, res_norms, ref_norms)
    "naive gradient descent method"
    # number of readouts point & image dimension
    nreads, image_dim = data.shape[-1], guess.shape[-1]
    # We start with an initial guess of zeros..
    xs = []
    x = guess
    # calculate b (lhs)
    b = mriop.operator_rhs(data)
    # residual norms
    gamma = mriop.operator_lhs(x) - b
    d = gamma
    res_norms = []
    # res_norms.append(frobenius_norm(gamma))
    # reference norms
    if ref is not None: ref_norms = []
    for i in range(iters):
        # Ad
        Ad = mriop.operator_lhs(d)
        # regularization
        Ad = Ad + lambd * d
        # alpha
        alpha = -np.vdot(gamma, d) / np.vdot(d, Ad)
        # update
        x = x + alpha * d
        xs.append(x)
        # new gamma
        gamma = mriop.operator_lhs(x) - b
        # residual
        delta = frobenius_norm(gamma)
        res_norms.append(delta)
        # reference norms
        if ref is not None: ref_norms.append(frobenius_norm(x-ref))
        if delta < tol:
            print("Converged after %i iterations to %1.6e" % (i + 1, delta))
            break
        if not np.mod(i, 1):
            print("Residuum at iter %i : %1.6e" % (i + 1, delta), end='\r')
        # beta
        beta = -np.vdot(gamma, Ad) / np.vdot(d, Ad)
        # new d
        d = gamma + beta * d
    xs = np.array(xs)
    if kfilter: xs = kspace_filter(xs, nreads, image_dim)
    if ref is not None:
        return xs if dstore else xs[-1], res_norms, ref_norms
    else:
        return xs if dstore else xs[-1], res_norms



"""`utils` module contains many useful _miscellaneous_ functions and utilities."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/99_utils.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/P4_HPD_cg/99_utils.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *



"""`io` module contains useful functions for loading and saving data."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/93_io.ipynb.

# %% auto 0
__all__ = ['epsilon', 'read_data']

# %% ../nbs/P4_HPD_cg/93_io.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/93_io.ipynb 4
import numpy as np
from prettytable import PrettyTable

# %% ../nbs/P4_HPD_cg/93_io.ipynb 5
from .plot import *
from .metrics import *
from .math import *

# %% ../nbs/P4_HPD_cg/93_io.ipynb 6
import h5py

# %% ../nbs/P4_HPD_cg/93_io.ipynb 7
epsilon = 1e-10

# %% ../nbs/P4_HPD_cg/93_io.ipynb 27
def read_data(
    fpath:str,  # path to the rawdata file
    R: int=1,   # acceleration factor
) -> tuple:     # output a tuple of ndarrays
    """
    Read the rawdata file and output tupple of np.ndarray of individual data
    Have a look at all the keys to see which individual data is available
    For the test brain data the individual data are rawdata, trajectory, InScale, and Coils    
    """
    data_dict = {} 
    
    with h5py.File(fpath, 'r') as f:
        keys = list(f.keys())
        print(keys)
        for key in keys:
            data_dict[key] = np.squeeze(f[key][()])
    
    rawdata = np.transpose(data_dict['rawdata'])
    trajectory = np.transpose(data_dict['trajectory'])
    
    # retrospectively undersampled
    rawdata = rawdata[:,::R,:]
    trajectory = trajectory[::R]
    
    return rawdata, trajectory, data_dict['InScale'], data_dict['Coils']



"""`plot` module contains many useful _plotting_ functions."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/90_plot.ipynb.

# %% auto 0
__all__ = ['epsilon', 'show_image_grid', 'show_compared_images', 'show_trajectory', 'gen_radial_traj']

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 4
import numpy as np
import matplotlib.pyplot as plt

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 5
plt.rcParams.update({'font.size': 13.25})
epsilon = 1e-10

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 7
def show_image_grid(
    images:np.ndarray,      # 3D-array (Nimgs, Nx, Ny) images   
    figsize:float=(10,5),   # size of the figure (hor, ver)
    rows:int=2,             # number of rows
    cols:int=5,             # number of columns
    cmap:str='gray',        # colormap type
    vmin:float=None,        # minimum value for the colorbar()
    vmax:float=None,        # maximum value for the colorbar()
    labels:str=None,        # (Nimgs)-size list of labels 
    title:str=None,         # title of the figure
    axis_on:bool=False,     # flag to turn the axis on/off
    colorbar_on:bool=True,  # flag to turn the colorbar on/off
    calc_flag:str='abs',    # flag to run abs, real, imag, or angle
    log_flag:str=None,      # flag to display as log or log10
    last_image:bool=False): # flag to only display the last image
    """
    Display multiple images in a (rows,cols) grid.
    """
    
    if labels == None:
        labels = ['#'+str(idx+1) for idx in (range(rows*cols))]
    
    images = np.squeeze(images)
    if images.ndim==2: images=images[np.newaxis]
    if images.shape[0]==1: last_image=True
    if last_image: 
        images = images[-1]
        images = images[np.newaxis]
        figsize = (10,10)
        labels = None
        rows = 1
        cols=1
    
    if images.shape[0]>rows*cols:
        print(f"Warning: number of images ({images.shape[0]}) is larger than number of panels ({rows}x{cols})!")    
    fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)
    if title == None: pass
    else: fig.suptitle(title)
    
    if calc_flag=='abs': images = np.abs(images)
    elif calc_flag=='real': images = np.real(images)
    elif calc_flag=='imag': images = np.imag(images)
    elif calc_flag=='angle': images = np.angle(images)
    
    if log_flag=='log10': images = np.log10(images)
    elif log_flag=='log': images = np.log(images)

    # Calculate the min and max values of all images
    if vmin is None: vmin = np.min(images)
    if vmax is None: vmax = np.max(images)
    
    if rows==1 and cols==1: idx_ax_iter = enumerate([axes])
    else: idx_ax_iter =  enumerate(axes.flatten())
    
    for i, ax in idx_ax_iter:
        if i < len(images):
            img = ax.imshow(images[i], cmap=cmap, vmin=vmin, vmax=vmax)
            if labels is not None: ax.set_title(labels[i])
            if axis_on==False: ax.axis('off')
        else:
            if axis_on==False: ax.axis('off')
    if colorbar_on:
        cax = fig.add_axes([0.95, 0.12, 0.025, 0.75])  # [left, bottom, width, height]
        plt.colorbar(img, cax=cax)
    else:
        plt.tight_layout()

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 16
def show_compared_images(
    X:np.ndarray,           # float-type 2D reference image
    Y:np.ndarray,           # float-type 2D compared image 
    figsize:float=(10,4),   # size of the figure (hor, ver)
    cmap:str='gray',        # colormap type
    vmin:float=None,        # minimum value for the colorbar()
    vmax:float=None,        # maximum value for the colorbar()
    labels:str=None,        # labels for X, Y, and X-Y
    title:str=None,         # title of the figure
    axis_on:bool=False,     # flag to turn the axis on/off
    colorbar_on:bool=True,  # flag to turn the colorbar on/off
    abs_calc:bool=False,     # flag to calculate abs or not
    diff_fac:float= 10):    # multiplication factor for (Y-X)
    """ Compare X and Y images by showing X, Y, and Y-X"""
    # concatinate individual image to images
    if abs_calc:
        images = np.stack((np.abs(X), np.abs(Y), diff_fac*np.abs(Y-X)))
    else:
        images = np.stack((X, Y, diff_fac*(Y-X)))
        
    # Calculate the min and max values of all images
    if vmin is None:
        vmin = np.min(images)
    if vmax is None:
        vmax = np.max(images)
    
    if labels == None:
        labels = ["X", "Y", "diff"+" (x"+str(diff_fac)+")"]
    else:
        labels[2] = labels[2] +" (x"+str(diff_fac)+")"
    
    rows, cols = 1,3
    fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=figsize)
    if title == None:
        pass
    else: 
        fig.suptitle(title)

    for i, ax in enumerate(axes.flatten()):
        if i < len(images):
            img = ax.imshow(images[i], cmap=cmap, vmin=vmin, vmax=vmax)
            ax.set_title(labels[i])
            if axis_on==False: ax.axis('off')
        else:
            if axis_on==False: ax.axis('off')
    if colorbar_on:
        cax = fig.add_axes([0.95, 0.15, 0.02, 0.7])  # [left, bottom, width, height]
        plt.colorbar(img, cax=cax)
    else:
        plt.tight_layout()

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 21
def show_trajectory(
    traj:np.ndarray,        # kspace trajectory shaped Nspoke x Nreadout x 3 (x,y,z)
    figsize:float=(10,10),   # size of the figure (hor, ver)
    golden:bool=False):     # flag to plot golden angle radial trajectory   
    """show radial kspace trajectory using plt.plot"""
    fig, ax = plt.subplots(figsize=figsize)
    num_spokes = traj.shape[0]
    colormap = plt.cm.plasma
    for ns in range(num_spokes):
        kx = np.squeeze(traj[ns,:,0])
        ky = np.squeeze(traj[ns,:,1])
        if golden: 
            ax.plot(kx, ky, color=colormap(ns/num_spokes), lw=1.5)
            ax.plot(kx, ky, color=colormap(ns/num_spokes), lw=1.5)
        else:
            angle = np.arctan2(ky[-1] - ky[0], kx[-1] - kx[0])
            ax.plot(kx, ky, color=colormap(angle/np.pi), lw=1.5)
    if golden:
        cbar = plt.colorbar(plt.cm.ScalarMappable(norm=plt.Normalize(0, num_spokes), cmap=colormap), ax=ax)
        cbar.set_label('Trajectory Number')
    else:
        cbar = plt.colorbar(plt.cm.ScalarMappable(norm=plt.Normalize(0, np.pi), cmap=colormap), ax=ax)
        cbar.set_label('Angle (radians)')

# %% ../nbs/P4_HPD_cg/90_plot.ipynb 23
def gen_radial_traj(
    coords:int=(64, 256),       # coordinates (#spokes, #readout)
    golden:bool=False,          # flag to generate golden angle trajectory
    full_spoke:bool=True):      # flag to generate full/half spoke 
    "generating radial trajectories"
    nspokes, nreads = coords
    traj = np.zeros((nspokes, nreads, 2))
    if full_spoke: 
        if golden: dtheta = np.pi*(111.25/180) # Li Feng et al
        else: dtheta = np.pi/nspokes
        idxs, radius = np.mgrid[0:nspokes, -0.5:0.5:1.0/nreads]
    else:
        if golden: dtheta = 2*np.pi*(137.51/360) # Li Feng et al
        else:dtheta = 2*np.pi/nspokes
        idxs, radius = np.mgrid[0:nspokes, 0.0:0.5:0.5/nreads]
    thetas = idxs * dtheta    
    traj[:, :, 0] = radius * np.cos(thetas) # x = r*cos(th)
    traj[:, :, 1] = radius * np.sin(thetas) # y = r*sin(th)
    return traj



"""`mri` module contains useful functions and classes for MRI forward and adjoint operations."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/94_mri.ipynb.

# %% auto 0
__all__ = ['kb_func', 'kb_kernel', 'calc_deapodizer', 'BaseOperator', 'NuFFT', 'hanning_window', 'coils_norm', 'coil_sens',
           'prewhiten', 'dcf', 'MriImagingModel', 'MriOperator', 'init_params', 'setup_params']

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 3
import h5py
import numpy as np
import scipy
import skimage
from matplotlib import pyplot as plt
import time
from abc import ABC, abstractmethod
import itertools

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 4
from .plot import *
from .metrics import *
from .math import *
from .io import *
import cgsense2023 as cgs2003

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 6
def kb_func(
  u:np.ndarray,   # kernel radius 
  width:int,      # kernel width 
  beta:float,     # scale params (Jackson '91.)
) -> np.ndarray:  # Kaiser-Bessel window function
    "Kaiser-Bessel window function y = I0 [beta*sqrt(1-(2u/w)^2)]/w"
    assert np.size(width) == 1
    x = beta * np.sqrt(1 - (2 * u / width) ** 2)
    y = scipy.special.i0(x) / width
    return y

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 8
def kb_kernel(
    kwidth:int,     # kernel width
    overgrid:float, # over-sampled factor 
    gridsize:int,   # grid size of over-sampled grid
    klength:int=32, # kernal look-up-table (lut) length
    ) -> tuple:     # return (kernel values and FT of kernel)
    "Calculate the appropriate Kaiser-Bessel gridding kernel and its FFT for deapodization."
    # From Beatty et al. equation [5]
    beta = np.pi*np.sqrt((kwidth/overgrid)**2 * (overgrid - 0.5) ** 2 - 0.8)
    # Kernel radii.
    u = np.linspace(0, (kwidth/2), int(np.ceil(klength*kwidth/2)))
    kvs = kb_func(u, kwidth, beta)
    kvs /= kvs.max()

    fft_y = np.flip(kvs) #TODO
    if np.mod(kwidth, 2): fft_y = np.concatenate((fft_y[:-1], kvs))
    else: fft_y = np.concatenate((fft_y, kvs))

    fft_y = np.abs(np.fft.fftshift(np.fft.ifft(fft_y, gridsize*klength))*fft_y.size)
    x = np.linspace(-int((gridsize//(2*overgrid))),int((gridsize//(2*overgrid)))-1,int((gridsize//(overgrid))))

    fft_y = fft_y[(fft_y.size / 2 + x).astype(int)]
    h = np.sinc(x / (gridsize * klength)) ** 2

    kvs_fft = fft_y * h
    kvs_fft /= kvs_fft.max()
    return kvs, kvs_fft

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 10
def calc_deapodizer(
    ktable_fft:np.ndarray,  # FFT of kaiser bessel func
) -> np.ndarray:            # deapodization func
    "calculate deapodizatoin func based on FFT of kaiser bessel func."
    deapodizer1D = 1.0 /ktable_fft
    deapodizer2D = np.outer(deapodizer1D, deapodizer1D)
    return deapodizer2D

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 17
class BaseOperator(ABC):
    "Abstract Base Class (ABC) for linear Operators."

    def __init__(self, data_par):
        "data_par (dict):(image_dim),(num_coils),(num_reads),(num_projs)"
        self.image_dim = data_par["image_dim"]
        self.grid_size = data_par["grid_size"]
        self.num_reads = data_par["num_reads"]
        self.num_coils = data_par["num_coils"]
        self.num_projs = data_par["num_projs"]

    @abstractmethod
    def forward(self, image):
        raise NotImplementedError

    @abstractmethod
    def adjoint(self, kdata):
        raise NotImplementedError

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 20
class NuFFT(BaseOperator):
    "Non-uniform FFT (NuFFT) object"
    def __init__(self, par, trajectory, fft_dim=(-2, -1)):
        super().__init__(par["Data"])
        
        self.trajectory = trajectory
        self.fft_dim = fft_dim
        self.overgrid = par["Data"]["overgrid"]
        self.klength = par["FFT"]["klength"]
        self.kwidth = par["FFT"]["kwidth"]
    
        self.ktable, ktable_FFT = kb_kernel(self.kwidth, self.overgrid, self.grid_size, self.klength)
        self.deapodizer = calc_deapodizer(ktable_FFT.astype(np.float64))
        
        self.num_kpoints = self.ktable.size    
        self.kwidth2 = par["FFT"]["kwidth"] / 2 #kernel radius i.e., half of the width
        
        if "density_cor" in par["FFT"].keys():
            self.density_cor = par["FFT"]["density_cor"]
        else:
            self.density_cor = np.ones(trajectory.shape[:-1], dtype=np.float64)
        
        if "grid_matrix" in par["FFT"].keys():
            self.grid_matrix = par["FFT"]["grid_matrix"]
        else:
            self.grid_matrix = self._generate_grid_matrix()
        
        self.grid_matrix_adj = self.grid_matrix.transpose()

    def adjoint(self, kdata):
        # "sqrt" density compensation
        denscor_inp = kdata * self.density_cor
        # gridding
        ogkspace = np.zeros((self.num_coils, self.grid_size, self.grid_size),dtype=np.complex64)
        for nc in range(self.num_coils):
            ogkspace[nc] = self.grid_matrix_adj.dot(denscor_inp[nc].flatten()).reshape(self.grid_size, self.grid_size)
        # iFFT
        image_grid = kspace2im_ifft2(ogkspace)
        # deapodization
        return self._deapodizer_adj(image_grid)

    def forward(self, image):
        # deapodization
        image_deapodizer = self._deapodizer_fwd(image)
        # FFT
        ogkspace = im2kspace_fft2(image_deapodizer)
        # Resample on radial trajectories
        kspace = np.zeros((self.num_coils, self.num_projs, self.num_reads),dtype=np.complex64)
        for nc in range(self.num_coils):
            kspace[nc] = self.grid_matrix.dot(ogkspace[nc].flatten()).reshape(self.num_projs, self.num_reads)
        # "sqrt" density compensation
        return kspace*self.density_cor

    def _deapodizer_adj(self, image_grid):
        # crop the grid_size_image to desired size image then multiply with deapodizer
        return crop_im(image_grid,(self.image_dim, self.image_dim))*self.deapodizer

    def _deapodizer_fwd(self, image):
        # multiply the desired image with deapodizer then zeropad to the grid_size_image 
        return zero_pad(image * self.deapodizer, (self.grid_size, self.grid_size))

    def _grid_lut(self, s, return_mapping=False):
        gridcenter = self.grid_size // 2
        sg = np.zeros((self.num_coils,self.grid_size,self.grid_size),dtype=np.complex64)
        grid_point_mapping = []
        kdat = s * self.density_cor
        for iproj, iread in itertools.product(range(self.num_projs),range(self.num_reads)):
            temp_mapping = []
            temp_mapping.append((iread, iproj))
            kx = self.trajectory[iproj, iread, 1]*self.overgrid
            ky = self.trajectory[iproj, iread, 0]*self.overgrid

            ixmin = int((kx - self.kwidth2) + gridcenter)
            ixmax = int((kx + self.kwidth2) + gridcenter) + 1
            iymin = int((ky - self.kwidth2) + gridcenter)
            iymax = int((ky + self.kwidth2) + gridcenter) + 1

            for gcount1 in np.arange(ixmin, ixmax+1):
                dkx = (gcount1 - gridcenter) - kx

                for gcount2 in np.arange(iymin, iymax+1):
                    dky = (gcount2 - gridcenter) - ky
                    dk = np.sqrt(dkx ** 2 + dky ** 2)

                    if dk < self.kwidth2:
                        fracind = dk / self.kwidth2 * (self.num_kpoints - 1)
                        kidx = int(fracind)
                        fracdk = fracind - kidx

                        kern = (self.ktable[kidx] * (1 - fracdk) + self.ktable[kidx + 1] * fracdk)

                        indx = gcount1
                        indy = gcount2

                        indx = np.mod(indx, self.grid_size)
                        indy = np.mod(indy, self.grid_size)

                        temp_mapping.append((indx, indy))

                        sg[:, indy, indx] += (kern * kdat[:,iproj,iread])
            
            grid_point_mapping.append(temp_mapping)

        if return_mapping:
            return sg, grid_point_mapping
        else:
            return sg

    def _invgrid_lut(self, sg):
        gridcenter = self.grid_size / 2

        s = np.zeros((self.num_coils,self.num_projs,self.num_reads),dtype=np.complex64)

        for iproj, ismpl in itertools.product(range(self.num_projs),range(self.num_reads)):

            kx = self.trajectory[iproj, ismpl, 1]*self.overgrid
            ky = self.trajectory[iproj, ismpl, 0]*self.overgrid

            ixmin = int((kx - self.kwidth2) + gridcenter)
            ixmax = int((kx + self.kwidth2) + gridcenter) + 1
            iymin = int((ky - self.kwidth2) + gridcenter)
            iymax = int((ky + self.kwidth2) + gridcenter) + 1

            for gcount1 in np.arange(ixmin, ixmax+1):
                dkx = (gcount1 - gridcenter) - kx

                for gcount2 in np.arange(iymin, iymax+1):
                    dky = (gcount2 - gridcenter) - ky
                    dk = np.sqrt(dkx ** 2 + dky ** 2)
                    if dk < self.kwidth2:

                        fracind = dk / self.kwidth2 * (self.num_kpoints - 1)
                        kidx = int(fracind)
                        fracdk = fracind - kidx

                        kern = (self.ktable[kidx] * (1 - fracdk) + self.ktable[kidx + 1] * fracdk)
                        indx = gcount1
                        indy = gcount2

                        if gcount1 < 0:
                            indx += self.grid_size
                            indy = self.grid_size - indy

                        if gcount1 >= self.grid_size:
                            indx -= self.grid_size
                            indy = self.grid_size - indy

                        if gcount2 < 0:
                            indy += self.grid_size
                            indx = self.grid_size - indx

                        if gcount2 >= self.grid_size:
                            indy -= self.grid_size
                            indx = self.grid_size - indx

                        s[:, iproj, ismpl] += kern*sg[:, indy, indx]
        return s * self.density_cor

    def _generate_grid_matrix(self):
        gridcenter = self.grid_size / 2
        rowind, colind, value = [], [], []
        for iproj, ismpl in itertools.product(range(self.num_projs),range(self.num_reads)):

            kx = self.trajectory[iproj, ismpl, 1]*self.overgrid
            ky = self.trajectory[iproj, ismpl, 0]*self.overgrid

            ixmin = int((kx - self.kwidth2) + gridcenter)
            ixmax = int((kx + self.kwidth2) + gridcenter) + 1
            iymin = int((ky - self.kwidth2) + gridcenter)
            iymax = int((ky + self.kwidth2) + gridcenter) + 1

            for gcount1 in np.arange(ixmin, ixmax+1):
                dkx = (gcount1 - gridcenter) - kx
                for gcount2 in np.arange(iymin, iymax+1):
                    dky = (gcount2 - gridcenter) - ky
                    dk = np.sqrt(dkx ** 2 + dky ** 2)
                    if dk < self.kwidth2:
                        fracind = (dk / self.kwidth2 * (self.num_kpoints - 1))
                        kidx = int(fracind)
                        fracdk = fracind - kidx
                        kern = (self.ktable[kidx] * (1 - fracdk) + self.ktable[kidx + 1] * fracdk)
                        indx = gcount1
                        indy = gcount2

                        if gcount1 < 0:
                            indx += self.grid_size
                            indy = self.grid_size - indy
                        if gcount1 >= self.grid_size:
                            indx -= self.grid_size
                            indy = self.grid_size - indy
                        if gcount2 < 0:
                            indy += self.grid_size
                            indx = self.grid_size - indx
                        if gcount2 >= self.grid_size:
                            indy -= self.grid_size
                            indx = self.grid_size - indx

                        rowind.append(iproj*self.num_reads+ismpl)
                        colind.append(indy*self.grid_size+indx)
                        value.append(kern)

        gridmat = scipy.sparse.coo_matrix((value, (rowind, colind)), shape=(self.num_projs*self.num_reads, self.grid_size**2))
        gridmat = gridmat.tocsc()
        return gridmat

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 27
def hanning_window(
    N:int=64,           # window size
) -> np.ndarray:    # window function values
    "1D symetric hanning window function"
    return 0.5*(1 - np.cos(2*np.pi*np.arange(N)/(N - 1)))

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 30
def coils_norm(
    coils:np.ndarray,   # raw coil sensitivities 
) -> tuple:             # tuple of (normalized coils, intensity_cor)
    "coil sensitivities normalization"
    # sosqrt = np.sqrt(np.sum(np.abs(coils)**2, axis=0)) # np.sum(np.sqrt((coils * np.conj(coils))),0)
    sosqrt = np.sum(np.sqrt((coils * np.conj(coils))),0)
    sosqrt[sosqrt == 0] = 1 
    intensity_cor = sosqrt # intensity correction
    return coils/sosqrt, intensity_cor

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 32
def coil_sens(
    kdata:np.ndarray,   # raw-uncompensated kspace
    traj:np.ndarray,    # kspace trajectories
    par:dict,           # dictionary contains paramss
) -> tuple:             # tuple of (coils, mask, intensity_cor)
    "Estimate complex coil sensitivities using SOS"
    FFT = NuFFT(par=par, trajectory=traj)
    wsize = 64 # size of selected low-res portion of kdata
    nspokes, nreads = kdata.shape[-2:]
    kdata_c = crop_im(kdata, (nspokes, wsize))
    hwin = hanning_window(N=wsize)
    kdata_cw = kdata_c*hwin
    kdata_w = zero_pad(kdata_cw, (nspokes,nreads))
    coil_images = FFT.adjoint(kdata_w * par["FFT"]["density_cor"])
    rss_image = rss_rec(coil_images)
    coils = coil_images/rss_image
    thresh = skimage.filters.threshold_otsu(rss_image)
    mask = rss_image > thresh*0.3
    mask = scipy.ndimage.binary_dilation(mask, iterations=10)
    coils, intensity_cor = coils_norm(coils)
    return coils, mask, intensity_cor

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 35
def prewhiten(
    kdata:np.ndarray,       # kspace data (Nc,Nsp,Nr)
    noise:np.ndarray,       # noises (Nc,Nsp,Nr)
    coils:np.ndarray=None,  # sensitivity maps (Nc,Nsp,Nr)
) -> tuple:                 # pre-whitened (kspace,coils)
    "Noise pre-whitenning"
    
    cov = np.cov(noise.reshape(noise.shape[0], -1))
    L = np.linalg.cholesky(cov)
    
    kdata_rsh = kdata.reshape([kdata.shape[0], -1])
    kdata_pwh = np.reshape(np.linalg.solve(L, kdata_rsh), kdata.shape)

    if coils is not None:
        coils_rsh = coils.reshape(coils.shape[0], -1)
        coils_pwh = np.reshape(np.linalg.solve(L, coils_rsh), coils.shape)
    
    return kdata_pwh, coils_pwh

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 38
def dcf(
    grid_matrix:np.ndarray, # gridding matrix
    num_projs:int=96,       # number of radial spokes
    num_reads:int=512,      # number of readout samples
) -> np.ndarray:            # density compensation function
    "density compensation function (DCF)"
    dens = grid_matrix.transpose()@(np.ones(grid_matrix.shape[0]))
    dens = 1/(grid_matrix@dens)
    dens = np.reshape(dens,(num_projs,num_reads))
    # sqrt because of NUFFT's Forward and Adjoint are adjoint with each other
    return np.sqrt(dens.astype(np.float32)) 

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 41
class MriImagingModel(BaseOperator):
    "The MRI imaging model including Coils."

    def __init__(self, par):
        super().__init__(par["Data"])

        self.NuFFT = NuFFT(par=par, trajectory=par["Data"]["trajectory"])
        self.coils = par["Data"]["coils"]
        self.conj_coils = np.conj(self.coils)

    def adjoint(self, kdata):
        return np.sum(self.NuFFT.adjoint(kdata) * self.conj_coils, 0)

    def forward(self, image):
        return self.NuFFT.forward(image * self.coils)

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 46
class MriOperator:
    "MRI Operator Class"
    def __init__(self, data_par, optimizer_par):
        self.image_dim = data_par["image_dim"]
        self.num_coils = data_par["num_coils"]
        self.mask = data_par["mask"]
        self.do_incor = data_par["do_intensity_cor"]
        self.incor = data_par["intensity_cor"].astype(np.complex64)
        self.maxit = optimizer_par["max_iter"]
        self.lambd = optimizer_par["lambda"]
        self.tol = optimizer_par["tolerance"]
        self.operator = None

    def set_operator(self, op):
        "Mandatory to set an operator (MRImagingOperator) prior to minimization."
        self.operator = op

    def operator_lhs(self, image):
        "Compute the LHS (A^T A x) of the normal equation."
        assert self.operator is not None, \
            "Please set an operator with the set_operation method"
        return self.operator_rhs(self.operator.forward(image))

    def operator_rhs(self, kspace):
        "Compute the RHS (A^T b) of the normal equation."
        assert self.operator is not None, \
            "Please set an operator with the set_operation method"
        return self.operator.adjoint(kspace)

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 56
def init_params()->dict: 
    "initialize dictionary of parameters"
    params = {}
    params['Optimizer'] = {} 
    params['FFT'] = {} 
    params['Data'] = {} 
    return params

# %% ../nbs/P4_HPD_cg/94_mri.ipynb 58
def setup_params(
    fpath:str,  # path to the data file
    R:int,      # acceleration factor (1,2,3,...)
)->dict:        # return dictionary contains all params and Data
    "One-stop-shop for setting up Parameters and Data"
    # Read input data
    kdata, trajectory, InScale, coils = read_data(fpath, R)
    noise = None
    
    # extract parameters    
    ncoils, nspokes, nreads = kdata.shape

    params = init_params()
    
    # Parameters: Optimizer
    params['Optimizer']['tolerance'] = 0
    params['Optimizer']['lambda'] = 0
    params['Optimizer']['max_iter'] = 10
    
    # Parameters: FFT
    params['FFT']['kwidth'] = 5
    params['FFT']['klength'] = 10_000
    
    # Parameters: Data
    params['Data']['noise'] = noise
    params['Data']['do_intensity_cor'] = True
    params['Data']['do_density_cor'] = True
    params['Data']['overgrid'] = 1.7033398310591292
    params['Data']['image_dim'] = 300
    params['Data']['mask'] = 1
    params['Data']['num_coils'] = ncoils
    params['Data']['num_reads'] = nreads
    params['Data']['num_projs'] = nspokes
    params['Data']['grid_size'] = 512
    params['Data']['intensity_cor'] = np.array([1.0])
    
    
    FFT = NuFFT(par=params, trajectory=trajectory)
    grid_matrix = FFT.grid_matrix
    params['FFT']['grid_matrix'] = grid_matrix
    density_cor = dcf(grid_matrix, num_projs=nspokes, num_reads=nreads)
    params['FFT']['density_cor'] = density_cor
    
    ## calculated mask and intensity correction - similar results except the scaling differences
    # _, mask, intensity_cor = coil_sens(kdata, trajectory, params)
    # params['Data']['intensity_cor'] = intensity_cor #np.array([1.0])
    # params['Data']['mask'] = mask
    
    # noise pre-whitening if noise is not None
    noise = None
    if noise is not None:
        kdata, coils = prewhiten(kdata, noise, coils)
    
    # coil sensitivity estimation if coils = None
    # coils = None
    if coils is None:
        coils, mask, intensity_cor = coil_sens(kdata, trajectory, params)
            
    params['Data']['coils'] = coils    
    # coils = params['Data']['coils']
    params['Data']['rawdata'] = kdata
    params['Data']['trajectory'] = trajectory
    params['Data']['rawdata_density_cor'] = kdata * density_cor
    return params



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/cgsense2023',
                'doc_host': 'https://hdocmsu.github.io',
                'git_url': 'https://github.com/hdocmsu/cgsense2023/',
                'lib_path': 'cgsense2023'},
  'syms': { 'cgsense2023.io': {'cgsense2023.io.read_data': ('P4_HPD_cg/io.html#read_data', 'cgsense2023/io.py')},
            'cgsense2023.math': { 'cgsense2023.math.crop_im': ('P4_HPD_cg/math.html#crop_im', 'cgsense2023/math.py'),
                                  'cgsense2023.math.im2kspace_fft2': ('P4_HPD_cg/math.html#im2kspace_fft2', 'cgsense2023/math.py'),
                                  'cgsense2023.math.kspace2im_ifft2': ('P4_HPD_cg/math.html#kspace2im_ifft2', 'cgsense2023/math.py'),
                                  'cgsense2023.math.kspace_filter': ('P4_HPD_cg/math.html#kspace_filter', 'cgsense2023/math.py'),
                                  'cgsense2023.math.radial_mask': ('P4_HPD_cg/math.html#radial_mask', 'cgsense2023/math.py'),
                                  'cgsense2023.math.rss_rec': ('P4_HPD_cg/math.html#rss_rec', 'cgsense2023/math.py'),
                                  'cgsense2023.math.zero_pad': ('P4_HPD_cg/math.html#zero_pad', 'cgsense2023/math.py')},
            'cgsense2023.metrics': { 'cgsense2023.metrics.mse': ('P4_HPD_cg/metrics.html#mse', 'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.nmse': ('P4_HPD_cg/metrics.html#nmse', 'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.nrmse': ('P4_HPD_cg/metrics.html#nrmse', 'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.print_metrics': ( 'P4_HPD_cg/metrics.html#print_metrics',
                                                                            'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.psnr': ('P4_HPD_cg/metrics.html#psnr', 'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.rmse': ('P4_HPD_cg/metrics.html#rmse', 'cgsense2023/metrics.py'),
                                     'cgsense2023.metrics.ssim': ('P4_HPD_cg/metrics.html#ssim', 'cgsense2023/metrics.py')},
            'cgsense2023.mri': { 'cgsense2023.mri.BaseOperator': ('P4_HPD_cg/mri.html#baseoperator', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.BaseOperator.__init__': ( 'P4_HPD_cg/mri.html#baseoperator.__init__',
                                                                            'cgsense2023/mri.py'),
                                 'cgsense2023.mri.BaseOperator.adjoint': ('P4_HPD_cg/mri.html#baseoperator.adjoint', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.BaseOperator.forward': ('P4_HPD_cg/mri.html#baseoperator.forward', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriImagingModel': ('P4_HPD_cg/mri.html#mriimagingmodel', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriImagingModel.__init__': ( 'P4_HPD_cg/mri.html#mriimagingmodel.__init__',
                                                                               'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriImagingModel.adjoint': ( 'P4_HPD_cg/mri.html#mriimagingmodel.adjoint',
                                                                              'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriImagingModel.forward': ( 'P4_HPD_cg/mri.html#mriimagingmodel.forward',
                                                                              'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriOperator': ('P4_HPD_cg/mri.html#mrioperator', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriOperator.__init__': ('P4_HPD_cg/mri.html#mrioperator.__init__', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriOperator.operator_lhs': ( 'P4_HPD_cg/mri.html#mrioperator.operator_lhs',
                                                                               'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriOperator.operator_rhs': ( 'P4_HPD_cg/mri.html#mrioperator.operator_rhs',
                                                                               'cgsense2023/mri.py'),
                                 'cgsense2023.mri.MriOperator.set_operator': ( 'P4_HPD_cg/mri.html#mrioperator.set_operator',
                                                                               'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT': ('P4_HPD_cg/mri.html#nufft', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT.__init__': ('P4_HPD_cg/mri.html#nufft.__init__', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT._deapodizer_adj': ( 'P4_HPD_cg/mri.html#nufft._deapodizer_adj',
                                                                            'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT._deapodizer_fwd': ( 'P4_HPD_cg/mri.html#nufft._deapodizer_fwd',
                                                                            'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT._generate_grid_matrix': ( 'P4_HPD_cg/mri.html#nufft._generate_grid_matrix',
                                                                                  'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT._grid_lut': ('P4_HPD_cg/mri.html#nufft._grid_lut', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT._invgrid_lut': ('P4_HPD_cg/mri.html#nufft._invgrid_lut', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT.adjoint': ('P4_HPD_cg/mri.html#nufft.adjoint', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.NuFFT.forward': ('P4_HPD_cg/mri.html#nufft.forward', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.calc_deapodizer': ('P4_HPD_cg/mri.html#calc_deapodizer', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.coil_sens': ('P4_HPD_cg/mri.html#coil_sens', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.coils_norm': ('P4_HPD_cg/mri.html#coils_norm', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.dcf': ('P4_HPD_cg/mri.html#dcf', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.hanning_window': ('P4_HPD_cg/mri.html#hanning_window', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.init_params': ('P4_HPD_cg/mri.html#init_params', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.kb_func': ('P4_HPD_cg/mri.html#kb_func', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.kb_kernel': ('P4_HPD_cg/mri.html#kb_kernel', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.prewhiten': ('P4_HPD_cg/mri.html#prewhiten', 'cgsense2023/mri.py'),
                                 'cgsense2023.mri.setup_params': ('P4_HPD_cg/mri.html#setup_params', 'cgsense2023/mri.py')},
            'cgsense2023.optimizer': { 'cgsense2023.optimizer.conjugate_gradient': ( 'P4_HPD_cg/optim.html#conjugate_gradient',
                                                                                     'cgsense2023/optimizer.py'),
                                       'cgsense2023.optimizer.frobenius_norm': ( 'P4_HPD_cg/optim.html#frobenius_norm',
                                                                                 'cgsense2023/optimizer.py'),
                                       'cgsense2023.optimizer.naive_steepest_descent': ( 'P4_HPD_cg/optim.html#naive_steepest_descent',
                                                                                         'cgsense2023/optimizer.py'),
                                       'cgsense2023.optimizer.steepest_descent': ( 'P4_HPD_cg/optim.html#steepest_descent',
                                                                                   'cgsense2023/optimizer.py')},
            'cgsense2023.plot': { 'cgsense2023.plot.gen_radial_traj': ('P4_HPD_cg/plot.html#gen_radial_traj', 'cgsense2023/plot.py'),
                                  'cgsense2023.plot.show_compared_images': ( 'P4_HPD_cg/plot.html#show_compared_images',
                                                                             'cgsense2023/plot.py'),
                                  'cgsense2023.plot.show_image_grid': ('P4_HPD_cg/plot.html#show_image_grid', 'cgsense2023/plot.py'),
                                  'cgsense2023.plot.show_trajectory': ('P4_HPD_cg/plot.html#show_trajectory', 'cgsense2023/plot.py')},
            'cgsense2023.utils': {}}}



"""`math` module contains useful functions for image operations or mathematical calculations in general."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/P4_HPD_cg/92_math.ipynb.

# %% auto 0
__all__ = ['epsilon', 'im2kspace_fft2', 'kspace2im_ifft2', 'rss_rec', 'crop_im', 'zero_pad', 'radial_mask', 'kspace_filter']

# %% ../nbs/P4_HPD_cg/92_math.ipynb 2
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *

# %% ../nbs/P4_HPD_cg/92_math.ipynb 4
import numpy as np
from prettytable import PrettyTable

# %% ../nbs/P4_HPD_cg/92_math.ipynb 5
from .plot import *
from .metrics import *

# %% ../nbs/P4_HPD_cg/92_math.ipynb 6
from skimage.metrics import peak_signal_noise_ratio, structural_similarity

# %% ../nbs/P4_HPD_cg/92_math.ipynb 7
epsilon = 1e-10

# %% ../nbs/P4_HPD_cg/92_math.ipynb 8
def im2kspace_fft2(
    img : np.ndarray,   # input image  
    dims=(-2,-1),       # dims for FFT operations
    norm="ortho",       # normalization option
) -> np.ndarray:        # centered-2D FFT of input images
    "Centered-2D-FFT converting images to kspace data"
    return np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(img, dims),norm=norm),dims)

# %% ../nbs/P4_HPD_cg/92_math.ipynb 15
def kspace2im_ifft2(
    ksp : np.ndarray,   # input ksapce data  
    dims=(-2,-1),       # dims for iFFT operations
    norm="ortho",       # normalization option
) -> np.ndarray:        # centered-2D iFFT of input kspace data
    "Centered-2D-iFFT converting kspace data to images"
    return np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(ksp, dims),norm=norm),dims)

# %% ../nbs/P4_HPD_cg/92_math.ipynb 22
def rss_rec(
    ims: np.ndarray,    # coil-by-coil input images
    dim: int=(0),       # axis for RSOS operation
)-> np.ndarray:         # RSOS (root sum-of-squares coil-combined image)
    "root sum-of-squares (RSOS) coil-combined reconstruction"
    return np.array(np.sqrt(np.sum(np.abs(ims)**2, axis=dim)/ims.shape[dim]))

# %% ../nbs/P4_HPD_cg/92_math.ipynb 25
def crop_im(
    ims:np.ndarray,     # input images
    sizes:int=(300,300), # new number of rows,cols
) -> np.ndarray:        # cropped images to (new_row, new_col) size
    "crop images to desired sizes (new_row, new_col)"
    row, col = ims.shape[-2:]   # original dimensions
    nrow, ncol = sizes          # new dimentions
    if nrow > row or ncol > col:
        print(f"Warning: cropped sizes {nrow,ncol} are larger than original sizes {row,col}!")
        return ims
    row1, col1 = (row - nrow)//2, (col - ncol)//2 # starts
    row2, col2 = (row + nrow)//2, (col + ncol)//2 # stops
    return ims[...,row1:row2, col1:col2]

# %% ../nbs/P4_HPD_cg/92_math.ipynb 37
def zero_pad(
    ims:np.ndarray,     # input images
    sizes:int=(512,512),# new number of rows,cols
) -> np.ndarray:        # zero-padded images to (new_row, new_col) size
    "zero-pad images to desired sizes (new_row, new_col)"
    row, col = ims.shape[-2:]   # original dimensions
    nrow, ncol = sizes          # new dimentions
    if nrow < row or ncol < col:
        print(f"Warning: padded sizes {nrow,ncol} are smaller than original sizes {row,col}!")
        return ims
    ims_out = np.zeros(ims.shape[:-2]+sizes, dtype=ims.dtype)
    row1, col1 = (nrow - row)//2, (ncol - col)//2 # starts
    row2, col2 = (row + nrow)//2, (col + ncol)//2 # stops
    ims_out[...,row1:row2, col1:col2] = ims
    return ims_out

# %% ../nbs/P4_HPD_cg/92_math.ipynb 46
def radial_mask(
    Nk:int=256,     # kspace dimension assumed Nkx=Nky=Nk
) -> np.ndarray:    # return radially-binary mask and center point    
    "create radially-binary mask for kspace filltering operation"
    kpoints = np.linspace(-0.5, 0.5, Nk, endpoint=False)
    xx, yy = np.meshgrid(kpoints, kpoints)
    mask = np.sqrt(xx**2 + yy**2)
    mask = np.where(np.abs(mask) > 0.5, 1, 0)
    mask = ~mask.astype(bool)
    center = Nk // 2
    return mask, center

# %% ../nbs/P4_HPD_cg/92_math.ipynb 52
def kspace_filter(
    x:np.ndarray,   # input images
    nreads:int,     # number of readout points
    image_dim:int,  # output image dimension
)->np.ndarray:      # kspace-filtered output images
    "Kspace filtering by zeroing out points outside the acquired trajectory"
    
    # data_type = mriop.operator.DTYPE
    # Nk, Nx = mriop.operator.num_reads, mriop.image_dim
    if x.ndim == 2: x = x[np.newaxis]
    xout = np.zeros_like(x)
    for idx in range(x.shape[0]):
        kspace0 = np.zeros((nreads,nreads),dtype=np.complex128)
        gridmask, gridcenter = radial_mask(nreads)
        img0 = np.zeros_like(kspace0) #
        Nx_range =slice(int(gridcenter-image_dim//2),int(gridcenter+image_dim//2))
        Ny_range = Nx_range
        img0[...,Nx_range, Ny_range] = x[idx]
        kspace1 = im2kspace_fft2(img0)
        masked_kspace1 = kspace1*gridmask
        img1 = kspace2im_ifft2(masked_kspace1)
        xout[idx] = img1[...,Nx_range, Ny_range]
    return np.squeeze(xout)
